/**
  Comic Book Thumbnailer for KDE 4 v0.1
  Copyright (C) 2009 Harsh J <harsh@harshj.com>

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

// comiccreator.cpp

#include "comiccreator.h"

#include <kdemacros.h>
#include <kzip.h>
#include <ktar.h>
#include <kmimetype.h>
#include <kdebug.h>
#include <kstandarddirs.h>
#include <ktempdir.h>
#include <kprocess.h>

#if defined(Q_OS_WIN)
    #include <QtCore/QProcess>
#else
    #include <kptyprocess.h>
#endif

#include <memory.h>

#include <QtGui/QImage>
#include <QtCore/QFile>
#include <QtCore/QFileInfo>
#include <QtCore/QDebug>
#include <QtCore/QStringList>

#define KIO_THUMB 11371

extern "C"
{
    KDE_EXPORT ThumbCreator *new_creator() {
        return new ComicCreator;
    }
};

ComicCreator::ComicCreator() : mComicCover(0) {}

ComicCreator::~ComicCreator() { delete mComicCover; }

bool ComicCreator::create(const QString &path, int width, int height, QImage &img) {

    bool result(false);

    // Detect mime type.
    const KMimeType::Ptr mime = KMimeType::findByFileContent(path);

    if (mime->is("application/x-cbz") || mime->name() == "application/zip") {

        // ZIP archive.
        result = ComicCreator::extractArchiveImage(path, "zip");

    } else if (mime->is("application/x-cbr") || mime->name() == "application/x-rar") {

        // RAR archive.
        result = ComicCreator::extractRARImage(path);

    } else if (mime->is("application/x-cbt") || mime->name() == "application/x-gzip") {

        // TAR archive.
        result = ComicCreator::extractArchiveImage(path, "tar");

//     } else if (mime->is("application/x-cb7") || mime->name() == "application/x-7z") {
//
//         result = ComicCreator::extract7ZImage(path);
//
//     } else if (mime->is("application/x-cba") || mime->name() == "application/x-ace") {
//
//         result = ComicCreator::extractACEImage(path);

    } else {
        result = false;
    }

    if(!mComicCover || !result) {
        kDebug(KIO_THUMB)<<"Error creating the comic book thumbnail.";
        return false;
    }

    // Perform a smooth scale for the requested dimensions.
    img = mComicCover->scaled(width, height, Qt::KeepAspectRatio, Qt::SmoothTransformation);
    return result;
}

void ComicCreator::filterImages(QStringList &entries, const bool sensitive=true) {
    // Sort based on condition, then remove non-image entries.
    if(sensitive) {
        entries.sort();
    } else {
        QMap<QString, QString> entryMap;
        Q_FOREACH(const QString &entry, entries) {
            entryMap.insert(entry.toLower(), entry);
        }
        entries = entryMap.values();
    }

    Q_FOREACH(const QString &entry, entries) {
        if (!(entry.endsWith(".gif", Qt::CaseInsensitive) || entry.endsWith(".jpg", Qt::CaseInsensitive) || entry.endsWith(".jpeg", Qt::CaseInsensitive) || entry.endsWith(".png", Qt::CaseInsensitive))) {
                entries.removeOne(entry);
        }
    }
}

bool ComicCreator::extractArchiveImage(QString path, QString type) {
    /// Extracts the cover image out of the .cbz or .cbt file.

    const KArchiveDirectory *cArchiveDir;

    if(type=="zip") {
        // Open the ZIP archive.
        KZip *cZip = new KZip(path);
        if (!cZip->open(QIODevice::ReadOnly)) {
            return false;
        }
        // Get the archive's entry directory.
        cArchiveDir = cZip->directory();
        if (!cArchiveDir) {
            return false;
        }
    } else if (type=="tar") {
        KTar *cTar = new KTar(path);
        if (!cTar->open(QIODevice::ReadOnly)) {
            return false;
        }
        // Get the archive's entry directory.
        cArchiveDir = cTar->directory();
        if (!cArchiveDir) {
            return false;
        }
    } else {
        cArchiveDir = 0;
        return false;
    }

    QStringList entries;

    // Get and filter the entries from the archive.
    ComicCreator::getArchiveFileList(entries, QString(), cArchiveDir);
    ComicCreator::filterImages(entries, false);

    // Extract the cover file.
    const KArchiveFile *coverFile = static_cast<const KArchiveFile*>(cArchiveDir->entry(entries[0]));
    if (!coverFile) {
        return false;
    }

    // Load cover data into image.
    mComicCover = new QImage();
    mComicCover->loadFromData(coverFile->data());
    if (!mComicCover) {
        return false;
    }

    return true;
}



void ComicCreator::getArchiveFileList(QStringList &entries, const QString &prefix, const KArchiveDirectory* dir) {
    /// Recursively list all files in the ZIP archive into 'entries'.
    Q_FOREACH (const QString &entry, dir->entries()) {
        const KArchiveEntry *e = dir->entry(entry);
        if (e->isDirectory()) {
        ComicCreator::getArchiveFileList(entries, prefix + entry + "/", static_cast<const KArchiveDirectory*>(e));
        } else if (e->isFile()) {
            entries.append(prefix + entry);
        }
    }
}

bool ComicCreator::extractRARImage(QString path) {
    /// Extracts the cover image out of the .cbr file.

    QString unrarPath;

    // Check if unrar is available. Get its path in 'unrarPath'.
    bool available = ComicCreator::isUnrarAvailable(unrarPath);
    if(!available) {
        kDebug(KIO_THUMB)<<"A suitable version of unrar is not available. Exiting.";
        return false;
    }

    QStringList entries;

    // Get the files and filter the images out.
    ComicCreator::getRARFileList(entries, path, unrarPath);
    ComicCreator::filterImages(entries, false);

    // Clear previously used data arrays.
    mStdOut.clear();
    mStdErr.clear();

    // Extract the cover file alone. Use verbose paths.
    // unrar x -n<file> path/to/archive /path/to/temp
    KTempDir *cUnrarTempDir = new KTempDir();
    ComicCreator::startProcess(unrarPath, QStringList() << "x" << "-n" + entries[0] << path << cUnrarTempDir->name());

    // Open cover file from the temp directory.
    QFile coverFile(cUnrarTempDir->name() + entries[0]);
    if (!coverFile.open(QIODevice::ReadOnly)) {
        return false;
    }

    // Load cover file data into image.
    mComicCover = new QImage();
    mComicCover->loadFromData(coverFile.readAll());
    if (!mComicCover) {
        return false;
    }

    // Temp directory's served its purpose.
    cUnrarTempDir->unlink();
    delete cUnrarTempDir;

    return true;
}

void ComicCreator::getRARFileList(QStringList &entries, const QString path, const QString unrarPath) {
    /// Get a verbose unrar listing so we can extract a single file later.
    // CMD: unrar vb /path/to/archive
    ComicCreator::startProcess(unrarPath, QStringList() << "vb" << path);
    entries = QString::fromLocal8Bit(mStdOut).split("\n", QString::SkipEmptyParts);
}

bool ComicCreator::isUnrarAvailable(QString &unrarPath) {
    /// Check the standard paths to see if a suitable unrar is available.
    unrarPath = KStandardDirs::findExe("unrar");
    if(unrarPath.isEmpty()) {
        unrarPath = KStandardDirs::findExe("unrar-nonfree");
    }
    if(!unrarPath.isEmpty()) {
        QProcess proc;
        proc.start(unrarPath, QStringList() << "--version");
        proc.waitForFinished(-1);
        const QStringList lines = QString::fromLocal8Bit(proc.readAllStandardOutput()).split("\n", QString::SkipEmptyParts);
        if(!lines.isEmpty()) {
            if (lines.first().contains("freeware")) {
                return true;
            }
        }
    }
    return false;
}

void ComicCreator::readProcessOut() {
    /// Read all std::out data and store to the data array.
    if (!mProcess)
        return;

    mStdOut += mProcess->readAllStandardOutput();
}

void ComicCreator::readProcessErr() {
    /// Read available std:err data and kill process if there is any.
    if (!mProcess)
        return;

    mStdErr += mProcess->readAllStandardError();
    if (!mStdErr.isEmpty())
    {
        mProcess->kill();
        return;
    }
}

void ComicCreator::finishedProcess(int exitCode, QProcess::ExitStatus exitStatus) {
    /// Run when process finishes.
    Q_UNUSED(exitCode)
    if (mLoop)
    {
        mLoop->exit(exitStatus == QProcess::CrashExit ? 1 : 0);
    }
}

int ComicCreator::startProcess(const QString processPath, const QStringList args) {
    /// Run a process and store std::out, std::err data in their respective buffers.
    int ret = 0;

#if defined(Q_OS_WIN)
    mProcess = new QProcess(this);
#else
    mProcess = new KPtyProcess(this);
    mProcess->setOutputChannelMode(KProcess::SeparateChannels);
#endif

    connect(mProcess, SIGNAL(readyReadStandardOutput()), SLOT(readProcessOut()));
    connect(mProcess, SIGNAL(readyReadStandardError()), SLOT(readProcessErr()));
    connect(mProcess, SIGNAL(finished(int, QProcess::ExitStatus)), SLOT(finishedProcess(int, QProcess::ExitStatus)));

#if defined(Q_OS_WIN)
    mProcess->start(processPath, args, QIODevice::ReadWrite | QIODevice::Unbuffered);
    ret = mProcess->waitForFinished(-1) ? 0 : 1;
#else
    mProcess->setProgram(processPath, args);
    mProcess->setNextOpenMode(QIODevice::ReadWrite | QIODevice::Unbuffered);
    mProcess->start();
    QEventLoop loop;
    mLoop = &loop;
    ret = loop.exec(QEventLoop::WaitForMoreEvents);
    mLoop = 0;
#endif

    delete mProcess;
    mProcess = 0;

    return ret;
}

ThumbCreator::Flags ComicCreator::flags() const {
    return DrawFrame;
}

#include "comiccreator.moc"